#!/usr/bin/env python

# Copyright (c) 2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import copy
import string

"""
This file provides class to manage decision trees and
their associated rules.
"""



class Rule:
    """
    Rule objects represent a rule associated with a path in
    a decision tree.
    """

    def __init__(self,cond,csq,training=(0,0),test=(0,0)):
        self.conditions = cond
        self.consequences = csq
        self.training = training
        self.test = test
        self.window = tuple(map(lambda x,y: x+y,training,test))
        
    def __repr__(self):
        from cStringIO import StringIO
        st = StringIO()
        st.write("\n"+"-"*50+"\n")
        st.write("IF\n")
        for feature,value in self.conditions.items():
            st.write("\t"+feature+" == "+str(value)+"\n") 
        st.write("THEN\n")
        for feature,value in self.consequences.items():
            st.write("\t"+feature+" = "+str(value)+"\n")

        st.write("PERFORMANCES : \n")
        for perf in [self.training,self.test,self.window]:
            st.write("\t "+string.join(map(lambda x: str(x),perf),'/')+"\n")
        st.write("-"*50+"\n")
        return st.getvalue()
    

    def __eq__(self,rule):
        # First checks that all conditions in "self" are in "rule" too.
        fv_list = rule.conditions.items()
        for t in self.conditions.items():
            if t not in fv_list:
                return 0

        # Then checks that all conditions in "rule" are in "self" too.
        fv_list = self.conditions.items()
        for t in rule.conditions.items():
            if t not in fv_list:
                return 0

        # Then checks that all csq in "rule" are in "self" too.
        fv_list = self.consequences.items()
        for t in rule.consequences.items():
            if t not in fv_list:
                return 0

        # Then checks that all csq in "self" are in "rule" too.
        fv_list = rule.consequences.items()
        for t in self.consequences.items():
            if t not in fv_list:
                return 0
 
        return 1

    def __cmp__(self,rule):
        """
        This method was overridden to allow list sorting according
        to the rule's performances.
        """
        
        # Tests particular cases (rules doesn't match anything !!)
        # This should not arrive
        if rule.window[0] == 0:
            return 1
        if self.window[1] == 0:
            return -1

        
        self_rate = self.window[0] / (self.window[1]*1.)
        rule_rate = rule.window[0] / (rule.window[1]*1.)

        # else, returns a value according to their performance comparisons.
        if self_rate > rule_rate:
            return 1
        if self_rate < rule_rate:
            return -1

        # for same performances, the best rules is the one which covers
        # the most examples
        if self.window[1] > rule.window[1]:
            return 1
        if self.window[1] < rule.window[1]:
            return -1

        # They're equals
        return 0
    
    def match(self,example):
        """
        Tests if the rule matches a training_example
        returns :
            - None if the conditions don't match the example
            - 0 if the conditions match but the consequences don't. This
              means that the rule is 'false' for this example.
            - 1 if the rule totally matches the example.
        """

        for feature,value in self.conditions.items():
            try:
                if value != example.features[feature]:
                    return None
            except KeyError:
                return None

        for feature,value in self.consequences.items():
            try:
                if value != example.features[feature]:
                    return 0
            except KeyError:
                return 0
        return 1


    def compute_relevance(self,sample):
        
        nb_examples = len(sample)
        nb_matched = 0
        for example in sample:
            result = self.match(example)
            if result is None:
                nb_examples = nb_examples - 1
                # else, only increment nb_matched when result=1
            elif result:
                nb_matched = nb_matched + 1
        return (nb_matched,nb_examples)

class RulesManager:
    """
    This class is here to manage the list of rules existing in
    the base and the new ones created during the learning process
    """
    
    def __init__(self, rules=None):
        self.rules = rules or []
        
    def set_rules(self,rules):
        self.rules = rules
    
    def load_rules_from_base(self):
        raise NameError("The method load_rules_from_base must be implemented")

    def update_base(self):
        raise NameError("The method update_base must be implemented")


    def compute_rules_relevance(self,training_sample = None,test_sample = None,window_sample = None):
        """
        Computes the rules' statistics.
        """
        for rule in self.rules:
            if training_sample is not None:
                rule.training = rule.compute_relevance(training_sample)
            if test_sample is not None:
                rule.test = rule.compute_relevance(test_sample)
            if window_sample is not None:
                rule.window = rule.compute_relevance(window_sample)
            else:
                # (a,b),(c,d) -> (a+c,b+d)
                rule.window = tuple(map(lambda x,y: x+y,rule.test,rule.training))

        
    def optimize_rules(self,sample):
        """
        This method removes all the 'unused' preconditions of
        each rules.
        The 'sample' should represent the entire set of examples (test+training)
        because rule performances musn't decrease over ALL the examples.
        """

        self.unoptimized_rules = copy.deepcopy(self.rules)
        changed = 1

        # Try to optimize as long as we can
        while changed:
            changed = 0
            new_set = copy.deepcopy(self.rules)
            for i in range(len(new_set)):
                rule = new_set[i]

                # In some particular cases, not having this test could
                # cause undesired results (a rule with no conditions
                # might be considered better than the non-optimized one)
                if len(rule.conditions) <= 1:
                    continue
                feature_list = rule.conditions.keys()
                new_rules = []
                # remove a condition and store the new rule
                for feature in feature_list:
                    new_conditions = copy.deepcopy(rule.conditions)
                    del new_conditions[feature]
                    new_rule = Rule(new_conditions,rule.consequences,rule.training,rule.test)

                    # XXX Fixme -- Try to do this another way
                    new_rule.__class__ = rule.__class__
                    new_rule.window = new_rule.compute_relevance(sample)
                    new_rules.append(new_rule)
                    
                # Insert the old_rule in first position so that the
                # ch_b_rule method will assign the default values with it.
                new_rules.insert(0,rule)
                new_set[i] = self.choose_best_rule(new_rules)
                if not new_set[i] == self.rules[i]:
                    changed = 1
            self.rules = new_set


        
    def choose_best_rule(self,rules):
        """
        Chooses the best rule within the list given in args.
        """
        perf = None
        covered_cases = None
        ind = 0
        for i in range(len(rules)):
            rule = rules[i]
            try:
                rule_perf = rule.window[0]/(rule.window[1]*1.)
            # Exceptions for Division by Zero or malformed data
            except:
                rule_perf = 0
            
            if perf is None or perf < rule_perf:
                perf = rule_perf
                ind = i
                covered_cases = rule.window[1]
            elif perf == rule_perf:
                if rule.window[1] >= covered_cases:
                    perf = rule_perf
                    ind = i
                    covered_cases = rule.window[1]
        return rules[ind]

    def remove_redundancy(self):
        """
        Removes all 'doubled' rules
        """
        new_set = copy.deepcopy(self.rules)
        for rule in self.rules:
            while new_set.count(rule) > 1:
                new_set.pop(new_set.index(rule))

        # Sort and reverse it in order to have the best rules at the beginning.
        new_set.sort()
        new_set.reverse()
        self.rules = new_set
        
class DecisionTree:
    """
    DecisionTree is the basic class used to manage decision trees.
    A decision tree is defined by :
       - a labelled Node where the label should be the tested feature at
         this level
       - a set of branches which are a set of Nodes associated with a value
         for the tested feature.
    """
    
    def __init__(self,feature = None):
        
        # the 'label' of the Node. It should be a string representing
        # the feature that the node is testing.
        self.testedFeature = feature

        # Branches are stored in a dictionnary. A "key" is a value associated
        # to the tested feature. A "value" is the subtree associated to this key.
        self.branches = {}

        # XXX : remove me.
        self.rules = []

    def add_branch(self,branch,featureValue):

        if self.branches.has_key(featureValue):
            print "Error : The Node : ",testedFeature," already have the branch ",featureValue
        else:
            self.branches[featureValue] = branch


    def compute_rules(self,conditions_dict=None):
        conditions = conditions_dict or {}
        if self.is_leaf():
            csq = {self.testedFeature[0]:self.testedFeature[1]}
            r = MeetingRule(conditions,csq)
            return [r]
        else:
            rules = []
            for value,branch in self.branches.items():
                new_conditions = copy.deepcopy(conditions)
                new_conditions[self.testedFeature] = value
                rules.append(branch.compute_rules(new_conditions))
            return reduce(lambda x,y: x+y,rules)
    
    
    def is_leaf(self):
        return len(self.branches) == 0

        
